# -*- coding: utf-8 -*-
#
# Brian 2 documentation build configuration file, created by
# sphinx-quickstart on Tue Aug 14 14:57:16 2012.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os, shutil

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))


# Mock the scipy module
class Mock(object):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    @classmethod
    def __getattr__(cls, name):
        if name in ('__file__', '__path__'):
            return '/dev/null'
        elif name[0] == name[0].upper():
            mockType = type(name, (), {})
            mockType.__module__ = __name__
            return mockType
        else:
            return Mock()

MOCK_MODULES = ['brian', 'brian.hears',   # mock Brian1 for the bridge
                # Mock additional packages to reduce installation time on
                # readthedocs
                'Cython', 'Cython.Compiler', 'Cython.Build',
                'scipy']
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()

# Store the examples and tutorials directory in an environment variable,
# used by the "example finder" that is called as part of the autoclass documentation
os.environ['BRIAN2_DOCS_EXAMPLE_DIR'] = root_dir = os.path.abspath('../examples')
os.environ['BRIAN2_DOCS_TUTORIALS_DIR'] = root_dir = os.path.abspath('../tutorials')

if 'BRIAN2_DOCS_QUICK_REBUILD' not in os.environ:
    # -- Automatically generate the reference documentation ------------------------
    import brian2.sphinxext.generate_reference as generate_reference
    # first generate the reference documentation
    target_dir = './reference'
    if os.path.exists(target_dir):
        shutil.rmtree(target_dir)
    os.makedirs(target_dir)
    abs_root = os.path.abspath('../brian2')
    generate_reference.main(abs_root, exclude_dirs=['tests', 'sphinxext'],
                            exclude_files=['conftest.py'],
                            destdir=target_dir)
    
    # -- Automatically generate the examples documentation -------------------------
    import brian2.sphinxext.generate_examples as generate_examples
    target_dir = './examples'
    root_dir = os.path.abspath('../examples')
    generate_examples.main(root_dir, target_dir)

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.8'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.doctest',
              'sphinx.ext.intersphinx', 'sphinx.ext.todo',
              'sphinx.ext.coverage', 'sphinx.ext.mathjax',
              'sphinx.ext.viewcode',
              'brian2.sphinxext.briandoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.extlinks',
              'sphinx_tabs.tabs']

# Set readthedocs theme if available
try:
    import sphinx_rtd_theme
    extensions += ['sphinx_rtd_theme']
    html_theme = 'sphinx_rtd_theme'
except ImportError:
    pass

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Brian 2'
import datetime, time
copyright = "2012–{}, Brian authors".format(
    datetime.datetime.fromtimestamp(
        int(os.environ.get("SOURCE_DATE_EPOCH", time.time())),
        tz=datetime.timezone.utc
    ).year
)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
from importlib.metadata import version

release = version('brian2')
# The short X.Y.Z version.
version = ".".join(release.split(".")[:3])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = 'brianobj'

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['brian2.']

sphinx_tabs_disable_tab_closing = True

# -- Options for HTML output ---------------------------------------------------

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}


# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/brian-logo.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'Brian2doc'

# Suppress warnings about the mybinder badges
suppress_warnings = ['image.nonlocal_uri']

# -- Options for LaTeX output --------------------------------------------------
latex_engine = 'xelatex'

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'Brian2.tex', 'Brian 2 Documentation',
   'Brian authors', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'brian2', 'Brian 2 Documentation',
     ['Brian authors'], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'Brian2', 'Brian 2 Documentation',
   'Brian authors', 'Brian2', 'One line description of project.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'


intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy' : ('https://numpy.org/doc/stable', None),
    'scipy' : ('https://docs.scipy.org/doc/scipy', None),
    'sympy' : ('https://docs.sympy.org/dev/', None)
}

autodoc_default_options = {'show-inheritance': True}

doctest_global_setup = 'from brian2 import *'
highlight_language = 'python'  # instead of python3 (default for sphinx>=1.4)

# Configure linking to github
import sphinx
sphinx_version = tuple(int(x) for x in sphinx.__version__.split('.'))
if sphinx_version >= (4, 0, 0): 
    extlinks = {'issue': ('https://github.com/brian-team/brian2/issues/%s',
                          '# %s')}
else:
    extlinks = {'issue': ('https://github.com/brian-team/brian2/issues/%s',
                          '# ')}
