"""
Package implementing the C++ "standalone" `Device` and `CodeObject`.
"""

from .codeobject import CPPStandaloneCodeObject
from .device import cpp_standalone_device
from .GSLcodeobject import GSLCPPStandaloneCodeObject
